[PATCH v5 07/28] x86/fpu/xstate: Calculate and remember dynamic xstate buffer sizes

From: Chang S. Bae
Date: Sun May 23 2021 - 15:38:56 EST


The xstate per-task buffer is currently embedded into struct fpu with
static size. To accommodate dynamic user xstates, record the maximum and
minimum buffer sizes.

Rename the size calculation function. It calculates the maximum xstate size
and sanity checks it with CPUID. It also calculates the static embedded
buffer size by excluding the dynamic user states from the maximum size.

Signed-off-by: Chang S. Bae <chang.seok.bae@xxxxxxxxx>
Reviewed-by: Len Brown <len.brown@xxxxxxxxx>
Cc: x86@xxxxxxxxxx
Cc: linux-kernel@xxxxxxxxxxxxxxx
---
Changes from v4:
* Massaged the function description, in preparation for the change
with a return value.

Changes from v3:
* Updated the changelog. (Borislav Petkov)
* Updated the code comment. (Borislav Petkov)
* Adjusted the calculation function naming.
* Moved out the new variable addition into a new patch.

Changes from v2:
* Updated the changelog with task->fpu removed. (Borislav Petkov)
* Renamed the in-line size variable.
* Updated some code comments.
---
arch/x86/kernel/fpu/xstate.c | 56 ++++++++++++++++++++++--------------
1 file changed, 35 insertions(+), 21 deletions(-)

diff --git a/arch/x86/kernel/fpu/xstate.c b/arch/x86/kernel/fpu/xstate.c
index 1d6613751625..74e0892bd8ec 100644
--- a/arch/x86/kernel/fpu/xstate.c
+++ b/arch/x86/kernel/fpu/xstate.c
@@ -654,24 +654,32 @@ static void check_xstate_against_struct(int nr)
}
}

-/*
- * This essentially double-checks what the cpu told us about
- * how large the XSAVE buffer needs to be. We are recalculating
- * it to be safe.
+/**
+ * calculate_xstate_sizes() - Calculate the xstate per-task buffer sizes -- maximum and minimum.
+ *
+ * Record the minimum and double-check the maximum against what the CPU told.
*
* Dynamic supervisor XSAVE features allocate their own buffers and are
* not covered by these checks. Only the size of the buffer for task->fpu
* is checked here.
*/
-static void do_extra_xstate_size_checks(void)
+static void calculate_xstate_sizes(void)
{
- int paranoid_xstate_size = FXSAVE_SIZE + XSAVE_HDR_SIZE;
+ int paranoid_min_size = FXSAVE_SIZE + XSAVE_HDR_SIZE;
+ int paranoid_max_size = FXSAVE_SIZE + XSAVE_HDR_SIZE;
int i;

for (i = FIRST_EXTENDED_XFEATURE; i < XFEATURE_MAX; i++) {
+ bool user_dynamic;
+
if (!xfeature_enabled(i))
continue;

+ user_dynamic = (xfeatures_mask_user_dynamic & BIT_ULL(i)) ? true : false;
+
check_xstate_against_struct(i);
/*
* Supervisor state components can be managed only by
@@ -681,23 +689,32 @@ static void do_extra_xstate_size_checks(void)
XSTATE_WARN_ON(xfeature_is_supervisor(i));

/* Align from the end of the previous feature */
- if (xfeature_is_aligned(i))
- paranoid_xstate_size = ALIGN(paranoid_xstate_size, 64);
+ if (xfeature_is_aligned(i)) {
+ paranoid_max_size = ALIGN(paranoid_max_size, 64);
+ if (!user_dynamic)
+ paranoid_min_size = ALIGN(paranoid_min_size, 64);
+ }
/*
* The offset of a given state in the non-compacted
* format is given to us in a CPUID leaf. We check
* them for being ordered (increasing offsets) in
* setup_xstate_features().
*/
- if (!using_compacted_format())
- paranoid_xstate_size = xfeature_uncompacted_offset(i);
+ if (!using_compacted_format()) {
+ paranoid_max_size = xfeature_uncompacted_offset(i);
+ if (!user_dynamic)
+ paranoid_min_size = xfeature_uncompacted_offset(i);
+ }
/*
* The compacted-format offset always depends on where
* the previous state ended.
*/
- paranoid_xstate_size += xfeature_size(i);
+ paranoid_max_size += xfeature_size(i);
+ if (!user_dynamic)
+ paranoid_min_size += xfeature_size(i);
}
- XSTATE_WARN_ON(paranoid_xstate_size != get_xstate_config(XSTATE_MAX_SIZE));
+ XSTATE_WARN_ON(paranoid_max_size != get_xstate_config(XSTATE_MAX_SIZE));
+ set_xstate_config(XSTATE_MIN_SIZE, paranoid_min_size);
}


@@ -798,14 +815,11 @@ static int __init init_xstate_size(void)
*/
set_xstate_config(XSTATE_MAX_SIZE, possible_xstate_size);

- /* Perform an extra check for the maximum size. */
- do_extra_xstate_size_checks();
-
/*
- * Set the minimum to be the same as the maximum. The dynamic
- * user states are not supported yet.
+ * Calculate and double-check the maximum size. Calculate and record
+ * the minimum size.
*/
- set_xstate_config(XSTATE_MIN_SIZE, possible_xstate_size);
+ calculate_xstate_sizes();

/* Ensure the minimum size fits in the statically-alocated buffer: */
if (!is_supported_xstate_size(get_xstate_config(XSTATE_MIN_SIZE)))
@@ -654,24 +654,32 @@ static void check_xstate_against_struct(int nr)
*
- * Dynamic supervisor XSAVE features allocate their own buffers and are
- * not covered by these checks. Only the size of the buffer for task->fpu
- * is checked here.
+ * Dynamic user states are stored in this per-task buffer. They account for the delta between the
+ * maximum and the minimum.
+ *
+ * Dynamic supervisor XSAVE features allocate their own buffers and are not covered by these checks.
+ *
+ * Return: nothing.
*/
--
2.17.1